"""main.py - Python basketball game written with Pygame
Copyright (c) 2015 Timothy Johnson <pythoneer@outlook.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import math
import os.path
import random

import pygame
from pygame.locals import *
try:
    import android
except ImportError:
    android = None
try:
    import pygame.mixer as mixer
except ImportError:
    import mixer

VERSION = "1.5.1"


class Basketball(pygame.sprite.Sprite):
    def __init__(self, game):
        super(Basketball, self).__init__()
        self.game = game
        self.hit_wall = False
        self.image = pygame.image.load(os.path.join("images", "ball.png")). \
            convert_alpha()
        self.radius = self.image.get_width() / 2
        self.sounds = [mixer.Sound(os.path.join("sounds", "bounce%d.ogg" %
                                                (i + 1))) for i in range(4)]
        self.speed = 2
        self.velocity = [0, 0]

    def set_initial_pos(self):
        self.action = None
        self.game.current_team = random.randint(0, 1)
        if self.game.current_team == 0:
            self.pos = [random.randint(60, 120),
                        random.randint(y - 120, y - 60)]
        else:
            self.pos = [random.randint(x - 120, x - 60),
                        random.randint(y - 120, y - 60)]

    def can_throw(self, check_action=True):
        not_auto = (self.game.current_team != self.game.auto_team or
                    self.game.level == 12)
        if check_action:
            return (not self.game.level_complete and
                    (not self.action or (self.action == "bounce" and
                                         not_auto)))
        else:
            return not self.game.level_complete and not_auto

    def move(self):
        if self.velocity[0]:
            self.pos[0] += self.velocity[0]
        if self.velocity[1]:
            self.pos[1] += self.velocity[1]

    def update(self):
        if self.action == "throw":
            if (self.radius < self.pos[0] < x - self.radius and
                    self.radius < self.pos[1] < y - self.radius):
                self.velocity[1] -= 0.001
                self.move()
                if (cx - 40 <= self.pos[0] <= cx + 40 and
                        70 <= self.pos[1] < 130 - self.radius):
                    self.end_throw(True)
                elif (not self.hit_wall and
                        (self.pos[0] < 60 or self.pos[0] > x - 60 or
                         (self.pos[1] < 60 and not
                          (cx - 128 < self.pos[0] < cx + 128)))):
                    self.game.sounds["wall"].play()
                    self.hit_wall = True
            else:
                self.end_throw(False)
        elif self.action == "basket":
            if self.pos[1] <= 247 + self.radius:
                if not self.velocity[0]:
                    self.pos[1] += self.velocity[1]
                elif (self.pos[0] <= cx - 55 + self.radius or
                        self.pos[0] >= cx + 55 - self.radius):
                    self.pos[0] += self.velocity[0]
                else:
                    self.move()
                    if (cx - abs(self.velocity[0]) < self.pos[0] < cx +
                            abs(self.velocity[0])):
                        self.velocity[0] = 0
            else:
                self.begin_land()
        elif self.action == "land":
            if self.pos[1] < self.end_pos[1]:
                self.velocity[1] += 0.001
                self.move()
            else:
                self.end_land()
        elif self.action == "bounce":
            if ((self.velocity[1] < 0 and self.pos[1] > self.bounce_height) or
                    (self.velocity[1] > 0 and self.pos[1] < self.end_pos[1])):
                if ((self.game.current_team == 0 and self.pos[0] <=
                        self.end_pos[0]) or (self.game.current_team == 1 and
                        self.pos[0] >= self.end_pos[0])):
                    self.velocity[0] = 0
                self.move()
            elif self.velocity[1] < 0:
                self.velocity[1] *= -1
            elif self.bounce_num < self.max_bounces:
                self.begin_bounce(self.bounce_num + 1)
            else:
                self.action = None

    def begin_throw(self, pos):
        if pos[1] < cy:
            self.game.sounds["what"].play()
            self.game.draw("Aim too close!")
            return
        elif pos[1] > self.pos[1]:
            self.game.sounds["what"].play()
            self.game.draw("Aim too low!")
            return
        pygame.mouse.set_cursor(*pygame.cursors.arrow)
        pygame.mouse.set_pos(cx, cy)
        theta = math.atan2(pos[1] - self.pos[1], pos[0] - self.pos[0])
        self.velocity[0] = self.speed * math.cos(theta)
        self.velocity[1] = self.speed * math.sin(theta)
        self.action = "throw"

    def end_throw(self, basket):
        if basket:
            self.begin_basket()
        else:
            if cx - 128 < self.pos[0] < cx + 128 and self.pos[1] < 151:
                self.game.sounds["rim"].play()
            self.hit_wall = False
            self.game.current_team ^= 1
            self.begin_land()

    def begin_basket(self):
        if self.game.current_team != self.game.auto_team:
            self.game.sounds["yay"].play()
            self.game.score += self.game.level + 2
        if (self.game.current_team == self.game.auto_team or
                self.game.level > 10):
            self.game.current_team ^= 1
        self.velocity[1] = self.speed * 0.75
        self.velocity[0] = (self.speed * (cx - self.pos[0]) /
                            abs(cx - self.pos[0]))
        self.action = "basket"

    def begin_land(self):
        self.end_pos = (random.randint(cx - self.radius, cx + self.radius),
                        random.randint(y - 120, y - 60))
        theta = math.atan2(self.end_pos[1] - self.pos[1],
                           self.end_pos[0] - self.pos[0])
        self.velocity[0] = self.speed * math.cos(theta)
        self.velocity[1] = self.speed * math.sin(theta)
        self.action = "land"

    def end_land(self):
        if self.game.current_team == 0:
            self.end_pos = (random.randint(60, 120),
                            random.randint(y - 120, y - 60))
        else:
            self.end_pos = (random.randint(x - 120, x - 60),
                            random.randint(y - 120, y - 60))
        if self.can_throw(False):
            pygame.mouse.set_cursor(*pygame.cursors.ball)
        self.max_bounces = random.randint(2, 4)
        self.begin_bounce(1)

    def begin_bounce(self, num):
        self.sounds[num - 1].play()
        self.bounce_height = int(self.pos[1] - y * 0.375 / num)
        self.bounce_num = num
        self.velocity[0] = (self.speed * (self.end_pos[0] - self.pos[0]) /
                            abs(self.end_pos[0] - self.pos[0]))
        self.velocity[1] = -self.speed
        self.action = "bounce"

    def auto_throw(self):
        theta = math.atan2(122 - self.radius - self.pos[1],
                           cx - self.pos[0])
        div = 12 + self.game.level
        theta = random.uniform(theta - math.pi / div, theta + math.pi / div)
        self.begin_throw((self.pos[0] + self.speed * math.cos(theta),
                          self.pos[1] + self.speed * math.sin(theta)))


class Game():
    def __init__(self, screen):
        self.do_pause = False
        self.mode = 0
        self.muted = False
        self.old_score = 0
        self.volume = 0.5

        self.background = None
        self.ball = Basketball(self)
        if not android:
            filename = os.path.join("fonts", "OpenSans-Regular.ttf")
        else:
            filename = None
        base_size = 20 if not android else 40
        self.fonts = [pygame.font.Font(filename, int(base_size * 0.8)),
                      pygame.font.Font(filename, base_size),
                      pygame.font.Font(filename, int(base_size * 1.25))]
        self.fonts[2].set_bold(True)
        self.screen = screen
        self.sounds = {"complete": mixer.Sound(os.path.join("sounds",
                                                            "complete.ogg")),
                       "fail": mixer.Sound(os.path.join("sounds", "fail.ogg")),
                       "rim": mixer.Sound(os.path.join("sounds", "rim.ogg")),
                       "wall": mixer.Sound(os.path.join("sounds", "wall.ogg")),
                       "what": mixer.Sound(os.path.join("sounds", "what.ogg")),
                       "yay": mixer.Sound(os.path.join("sounds", "yay.ogg"))}

        with open("credits.txt", 'r') as fileobj:
            self.credits = fileobj.readlines()
        self.credits_top = int(y * 0.75)

    def draw_intro(self):
        self.screen.fill((0, 0, 0))
        self.credits_top -= 1
        for i, line in enumerate(self.credits):
            text = self.fonts[0].render(line.rstrip(), 1, (255, 255, 255))
            textpos = text.get_rect(center=(cx, self.credits_top + i * 30))
            self.screen.blit(text, textpos)
        if self.credits_top + (i + 1) * 30 < y / 4:
            self.credits_top = int(y * 0.75)
        pygame.draw.rect(self.screen, (0, 0, 0), [0, 0, x, y / 4])
        text = self.fonts[2].render("PB-Ball", 1, (255, 128, 0))
        self.screen.blit(text, text.get_rect(center=(cx, y / 8)))
        pygame.draw.rect(self.screen, (0, 0, 0), [0, int(y * 0.75), x, y / 4])
        text = "CLICK TO START" if not android else "TAP TO START"
        text = self.fonts[1].render(text, 1, (0, 0, 255))
        textpos = text.get_rect(center=(cx, int(y * 0.875) -
                                        self.fonts[1].get_height() / 2))
        self.screen.blit(text, textpos)
        if not android:
            text = "press P to pause game"
        else:
            text = "press any button to pause game"
        text = self.fonts[0].render(text, 1, (0, 0, 255))
        textpos = text.get_rect(center=(cx, int(y * 0.875) +
                                        self.fonts[0].get_height() / 2))
        self.screen.blit(text, textpos)
        pygame.display.flip()

    def begin_game(self):
        self.level = 1
        self.level_up = False
        self.lives = 3
        self.score = self.old_score = 0
        if self.mode == 0:
            self.fonts[2].set_bold(False)
        self.begin_level()

    def begin_level(self):
        self.ball.set_initial_pos()
        self.auto_team = 1 - self.current_team
        if self.level_up:
            self.level += 1
            self.old_score = self.score
        self.start_time = pygame.time.get_ticks()
        if not android:
            text = "Click to aim the ball"
        else:
            text = "Tap to aim the ball"
        self.draw(text)
        mixer.music.play()
        mixer.music.set_volume(self.volume if not self.muted else 0)
        pygame.mouse.set_cursor(*pygame.cursors.ball)
        self.mode = 1

    @property
    def level_complete(self):
        if self.level < 11:
            new_score = (self.level + 2) * (self.level + 1)
        else:
            new_score = (self.level + 2) * 6
        return self.score >= self.old_score + new_score

    def draw(self, message=None):
        if not self.background:
            self.render_bg()
        self.screen.blit(self.background, (0, 0))
        if self.ball.action == "basket":
            self.screen.blit(self.ball.image,
                             [int(self.ball.pos[0] + 0.5) - self.ball.radius,
                              int(self.ball.pos[1] + 0.5) - self.ball.radius])
        pygame.draw.line(self.screen, (255, 128, 0), [cx - 55, 132],
                         [cx + 55, 132], 5)
        for i in range(12):
            dx = 37 if i < 9 else 37 + 4 * (i - 8)
            y2 = 247 if i < 9 else 247 - 27 * (i - 8)
            pygame.draw.line(self.screen, (247, 247, 247),
                             [cx - 54 + i * 9, 135], [cx - dx + i * 9, y2])
            dx = 62 if i > 2 else 62 - 4 * (3 - i)
            y2 = 247 if i > 2 else 247 - 27 * (3 - i)
            pygame.draw.line(self.screen, (247, 247, 247),
                             [cx - 45 + i * 9, 135], [cx - dx + i * 9, y2])
        if self.ball.action != "basket":
            self.screen.blit(self.ball.image,
                             [int(self.ball.pos[0] + 0.5) - self.ball.radius,
                              int(self.ball.pos[1] + 0.5) - self.ball.radius])
        self.screen.blit(self.fonts[1].render("Score: %d" % self.score, 1,
                                              (0, 0, 255)), [10, 5])
        self.screen.blit(self.fonts[0].render("Level: %d" % self.level, 1,
                                              (0, 0, 255)),
                         [10, self.fonts[1].get_height() + 5])
        time = max(0, int(60 - (pygame.time.get_ticks() - self.start_time) /
                          1000 + 0.5))
        self.draw_time(time, False)
        text = self.fonts[0].render("Lives: %d" % self.lives, 1, (255, 0, 0))
        self.screen.blit(text, [x - text.get_width() - 10,
                                self.fonts[1].get_height() + 5])
        if not message:
            baskets = (self.score - self.old_score) / (self.level + 2)
            baskets2 = self.level + 1 if self.level < 11 else 6
            message = "Baskets: %d/%d" % (baskets, baskets2)
        text = self.fonts[1].render(message, 1, (0, 255, 0))
        self.screen.blit(text, text.get_rect(center=(cx, y - 20)))
        pygame.display.flip()

    def render_bg(self):
        self.background = pygame.Surface((x, y))
        self.background.convert()
        y_floor = int(y * 0.625)
        pygame.draw.rect(self.background, (192, 192, 192), [0, 0, x, y_floor])
        pygame.draw.rect(self.background, (128, 128, 128),
                         [0, y_floor, x, y - y_floor])
        pygame.draw.rect(self.background, (0, 0, 0), [cx - 128, 5, 256, 146])
        pygame.draw.rect(self.background, (255, 255, 255),
                         [cx - 123, 10, 246, 136])
        pygame.draw.rect(self.background, (128, 0, 0), [cx - 40, 70, 80, 65])
        pygame.draw.rect(self.background, (255, 255, 255),
                         [cx - 35, 75, 70, 55])

    def draw_time(self, time, update=True):
        if time == 60:
            text = "Time: 1:00"
        else:
            text = "Time: 0:%s" % str(time).zfill(2)
        text = self.fonts[1].render(text, 1, (255, 0, 0))
        w, h = text.get_size()
        if update:
            pygame.draw.rect(self.screen, (192, 192, 192),
                             [x - w - 15, 5, w + 5, h])
        if not (time <= 5 and pygame.time.get_ticks() % 1000 >= 500):
            self.screen.blit(text, [x - w - 10, 5])
        if update:
            pygame.display.update([x - w - 10, 5, w, h])

    def toggle_pause(self):
        if self.mode == 1:
            self.pause_time = pygame.time.get_ticks()
            background = pygame.Surface((x, y), SRCALPHA)
            background.convert_alpha()
            background.fill((0, 0, 0, 128))
            self.screen.blit(background, (0, 0))
            text = self.fonts[2].render("Game Paused", 1, (0, 255, 0))
            textpos = text.get_rect(center=(cx, cy -
                                            self.fonts[2].get_height() / 2))
            self.screen.blit(text, textpos)
            text = "press P to resume" if not android else "tap to resume"
            text = self.fonts[1].render(text, 1, (0, 255, 0))
            textpos = text.get_rect(center=(cx, cy +
                                            self.fonts[1].get_height() / 2))
            self.screen.blit(text, textpos)
            pygame.display.flip()
            if mixer.music.get_busy():
                mixer.music.pause()
            pygame.mouse.set_cursor(*pygame.cursors.arrow)
            self.do_pause = False
            self.mode = 2
        elif self.mode == 2:
            self.start_time += pygame.time.get_ticks() - self.pause_time
            self.draw()
            if mixer.music.get_busy():
                mixer.music.unpause()
            if self.ball.can_throw(False):
                pygame.mouse.set_cursor(*pygame.cursors.ball)
            self.mode = 1
        elif self.mode == 3:
            self.do_pause = True

    def end_level(self, level_up=True):
        if mixer.music.get_busy():
            mixer.music.stop()
        if level_up:
            self.sounds["complete"].play()
            self.score += self.level * int(60 - (pygame.time.get_ticks() -
                                                 self.start_time) / 1000 + 0.5)
        else:
            self.sounds["fail"].play()
            self.lives -= 1
            self.score = self.old_score
        self.start_time = pygame.time.get_ticks()
        background = pygame.Surface((x, y), SRCALPHA)
        background.convert_alpha()
        background.fill((0, 0, 0, 128))
        self.screen.blit(background, (0, 0))
        if self.lives == -1 or (self.level == 12 and level_up):
            self.end_game()
            return
        elif level_up:
            text = "Level Complete"
        else:
            text = "Try Again"
        text = self.fonts[2].render(text, 1, (0, 255, 0))
        self.screen.blit(text, text.get_rect(center=(cx, cy)))
        pygame.display.flip()
        pygame.mouse.set_cursor(*pygame.cursors.arrow)
        self.level_up = level_up
        self.mode = 3

    def end_game(self):
        self.screen.fill((0, 0, 0))
        if self.level == 12 and self.level_up:
            text = "All Levels Complete"
        else:
            text = "Game Over"
        text = self.fonts[2].render(text, 1, (255, 0, 0))
        self.screen.blit(text, text.get_rect(center=(cx, y / 4)))
        highest = 0
        if os.path.isfile("score.dat"):
            with open("score.dat", 'r') as fileobj:
                highest = int(fileobj.read(), 2)
        if self.score > highest:
            with open("score.dat", 'w') as fileobj:
                fileobj.write(str(bin(self.score)))
            highest = self.score
        text = self.fonts[1].render("Score: %d" % self.score, 1, (0, 255, 0))
        textpos = text.get_rect(center=(cx, cy -
                                        self.fonts[1].get_height() / 2))
        self.screen.blit(text, textpos)
        text = self.fonts[1].render("Highest Score: %d" % highest, 1,
                                    (0, 255, 0))
        textpos = text.get_rect(center=(cx, cy +
                                        self.fonts[1].get_height() / 2))
        self.screen.blit(text, textpos)
        text = "CLICK TO REPLAY" if not android else "TAP TO REPLAY"
        text = self.fonts[1].render(text, 1, (0, 0, 255))
        self.screen.blit(text, text.get_rect(center=(cx, int(y * 0.75))))
        pygame.display.flip()
        pygame.mouse.set_cursor(*pygame.cursors.arrow)
        self.mode = 4


def main():
    pygame.init()
    mixer.init()
    icon = pygame.display.set_icon(pygame.image.load(os.path.join("images",
                                                                  "icon.png")))
    global x, y, cx, cy
    x = 1024
    y = 768 if not android else 576
    cx, cy = x / 2, y / 2
    screen = pygame.display.set_mode((x, y))
    screen.convert()
    caption = pygame.display.set_caption("PB-Ball %s" % VERSION)
    if android:
        android.init()
        android.map_key(android.KEYCODE_MENU, pygame.K_p)
        android.map_key(android.KEYCODE_BACK, pygame.K_p)
    game = Game(screen)
    mixer.music.load(os.path.join("sounds", "music.ogg"))
    clock = pygame.time.Clock()

    while True:
        if not (game.mode == 1 and game.ball.action):
            clock.tick(60)
        if android and android.check_pause():
            if game.mode == 1:
                game.toggle_pause()
            android.wait_for_resume()
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                return
            elif event.type == pygame.MOUSEBUTTONDOWN:
                if game.mode == 0 or game.mode == 4:
                    game.begin_game()
                elif game.mode == 1 and game.ball.can_throw():
                    game.ball.begin_throw(pygame.mouse.get_pos())
                elif android and game.mode == 2:
                    game.toggle_pause()
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_F4:
                    mixer.music.set_volume(0 if not game.muted else
                                           game.volume)
                    game.muted = not game.muted
                elif event.key == pygame.K_F9:
                    if not game.muted:
                        game.volume = max(0, mixer.music.get_volume() - 0.05)
                    else:
                        game.volume = max(0, game.volume - 0.05)
                    mixer.music.set_volume(game.volume)
                elif event.key == pygame.K_F10:
                    if not game.muted:
                        game.volume = min(mixer.music.get_volume() + 0.05, 1)
                    else:
                        game.volume = min(game.volume + 0.05, 1)
                    mixer.music.set_volume(game.volume)
                elif event.key == pygame.K_p:
                    if not android or game.mode != 2:
                        game.toggle_pause()
        if game.mode == 0:
            game.draw_intro()
        else:
            time = int((pygame.time.get_ticks() -
                        game.start_time) / 1000 + 0.5)
            if game.mode == 1:
                if time >= 60:
                    game.end_level(game.level_complete)
                elif game.ball.action:
                    game.ball.update()
                    game.draw()
                elif game.current_team == game.auto_team or game.level == 12:
                    game.ball.auto_throw()
                elif game.level_complete:
                    game.end_level()
                else:
                    game.draw_time(60 - time)
            elif game.mode == 3 and time > 2:
                game.begin_level()
                if game.do_pause:
                    game.toggle_pause()
            elif android:
                pygame.display.flip()
        fps = clock.get_fps()
        if fps > 2 and not (game.mode == 1 and game.ball.action):
            speed = 120 / fps if not android else 360 / fps
            game.ball.speed = min(speed, game.ball.speed + 1)


if __name__ == "__main__":
    main()
